use anyhow::Result;
use evdev_rs::{
    enums::{EventCode, EventType},
    Device as EvDevice, DeviceWrapper,
};
use std::{collections::HashMap, fs::File, os::unix::prelude::OsStrExt};

use crate::event_code_iters::*;
use crate::graph_types::{Descriptor, NodeDescriptor};

pub struct Device {
    device: EvDevice,
    path: String,

    supported_events: HashMap<EventType, Vec<EventCode>>,
}

impl Device {
    fn enumerate() -> Result<Vec<String>> {
        let mut files = Vec::new();

        if let Some(readdir) = std::fs::read_dir("/dev/input").ok() {
            for entry in readdir {
                let path = entry?.path();
                let fname = path.file_name().unwrap();
                if fname.as_bytes().starts_with(b"event") {
                    // Try to load read/write, then fall back to read-only.
                    if File::options()
                        .read(true)
                        .write(true)
                        .open(&path)
                        .or_else(|_| File::options().read(true).write(false).open(&path))
                        .is_ok()
                    {
                        files.push(path.as_path().to_str().unwrap().to_string());
                    }
                }
            }
        }

        Ok(files)
    }

    pub fn query() -> Result<Vec<Device>> {
        Self::enumerate()?
            .into_iter()
            .map(|path| {
                let file = File::open(&path)?;
                let device = EvDevice::new_from_file(file)?;
                let mut supported_events = HashMap::new();

                if device.has(&EventType::EV_KEY) {
                    let mut keys = Vec::new();

                    for key in ev_keys().iter() {
                        if device.has(key) {
                            keys.push(*key);
                        }
                    }

                    supported_events.insert(EventType::EV_KEY, keys);
                }

                if device.has(&EventType::EV_REL) {
                    let mut rels = Vec::new();

                    for rel in ev_rels().iter() {
                        if device.has(rel) {
                            rels.push(*rel);
                        }
                    }

                    supported_events.insert(EventType::EV_REL, rels);
                }

                if device.has(&EventType::EV_ABS) {
                    let mut abss = Vec::new();

                    for abs in ev_abss().iter() {
                        if device.has(abs) {
                            abss.push(*abs);
                        }
                    }

                    supported_events.insert(EventType::EV_ABS, abss);
                }

                if device.has(&EventType::EV_FF) {
                    let mut ffs = Vec::new();

                    for ff in ev_ffs().iter() {
                        if device.has(ff) {
                            ffs.push(*ff);
                        }
                    }

                    supported_events.insert(EventType::EV_FF, ffs);
                }

                if device.has(&EventType::EV_FF_STATUS) {
                    let mut ffss = Vec::new();

                    for ffs in ev_ffss().iter() {
                        if device.has(ffs) {
                            ffss.push(*ffs);
                        }
                    }

                    supported_events.insert(EventType::EV_FF_STATUS, ffss);
                }

                if device.has(&EventType::EV_REP) {
                    let mut reps = Vec::new();

                    for rep in ev_reps().iter() {
                        if device.has(rep) {
                            reps.push(*rep);
                        }
                    }

                    supported_events.insert(EventType::EV_REP, reps);
                }

                Ok(Device {
                    device,
                    path,

                    supported_events,
                })
            })
            .collect()
    }

    pub fn name(&self) -> &str {
        self.device.name().unwrap_or("Unnamed Device")
    }

    pub fn to_descriptor(&self) -> NodeDescriptor {
        let mut inputs = match self.supported_events.get(&EventType::EV_FF) {
            Some(ff) => ff
                .iter()
                .map(|kind| {
                    (
                        code_to_string(kind),
                        match self.device.abs_info(kind) {
                            Some(abs_info) => Descriptor::Slider {
                                min: abs_info.minimum,
                                max: abs_info.maximum,
                            },
                            None => Descriptor::Button,
                        },
                    )
                })
                .collect(),
            None => Vec::new(),
        };

        let outputs = Self::flatten(
            self.supported_events
                .iter()
                .map(|(event_type, kinds)| {
                    let t: Option<Vec<(String, Descriptor)>> = match *event_type {
                        EventType::EV_ABS => Some(
                            kinds
                                .iter()
                                .map(|kind| {
                                    (
                                        code_to_string(kind),
                                        match self.device.abs_info(kind) {
                                            Some(abs_info) => Descriptor::Slider {
                                                min: abs_info.minimum,
                                                max: abs_info.maximum,
                                            },
                                            None => Descriptor::Button,
                                        },
                                    )
                                })
                                .collect(),
                        ),
                        EventType::EV_REL => Some(
                            kinds
                                .iter()
                                .map(|kind| {
                                    (
                                        code_to_string(kind),
                                        match self.device.abs_info(kind) {
                                            Some(abs_info) => Descriptor::Slider {
                                                min: abs_info.minimum,
                                                max: abs_info.maximum,
                                            },
                                            None => Descriptor::Button,
                                        },
                                    )
                                })
                                .collect(),
                        ),
                        EventType::EV_KEY => Some(
                            kinds
                                .iter()
                                .map(|kind| (code_to_string(kind), Descriptor::Button))
                                .collect(),
                        ),

                        _ => None,
                    };

                    t
                })
                .filter_map(|f| f)
                .collect::<Vec<Vec<(String, Descriptor)>>>(),
        );

        inputs.insert(0, (format!("Path: {}", self.path), Descriptor::Info));

        NodeDescriptor {
            op_name: self.name().to_string(),
            inputs,
            outputs,
        }
    }

    fn flatten<T>(vv: Vec<Vec<T>>) -> Vec<T> {
        vv.into_iter().flatten().collect()
    }
}

impl std::fmt::Debug for Device {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("Device")
            .field("name", &self.device.name())
            .field("path", &self.path)
            .field("supported_events", &self.supported_events)
            .finish()
    }
}
